/*--------------------------------------------------------------------------*/
/** @file atmos.h
*                  MASS project: TURBINA module file                       
*                       header file for atmos.c                             
*                                                                          
* The module ATMOS specifies all the top-level calculations which are aimed to
* compute the atmospheric parameters and the turbulence profile. The weight
* functions, described themselves in the module WEIF, are selected and 
* initialized here. Also, the basis functions are produced from them in this
* module. With the help of these functions, the various integrals over the
* altitude which approximate the quantities in interest as seeing, isoplanatic
* or isokinetic angles etc are calculated and the low-resolution turbulence
* profile is computed.
*
* \b Usage:
*
* Prior to the scientific operations using the observation data, the module
* should be initialized with at_init(). Since the module acts as an interface
* to the module "weif", some parameters for weight initialization (which are not
* changed frequently) are transmitted via at_init(), some others (such as number
* of measurements or spectral energy distribution - via at_update(). Note, that 
* the initialization of the module "scind" have to be done separately. Each time
* some basic information has to be changed (apertures geometry, transmission
* function, altitude grid), the module has to be shut down with at_done() and
* reinitialized with at_init() again.

* Each time the new object is selected, the weight functions must be updated
* using the object's own spectral energy distribution (SED) file via the call to
* at_update() function. 

* Before starting the "mode", i.e. the planned measurement
* number is known (base times per accumulation time), the function at_alloc() must
* be called \b before the first calculation of atmospheric integrals. at_alloc()
* requires this guessed index measurement number per accumulation time.
* It must not be less than actual to avoid the storages overfull.

* Then, after some (instantaneous) indices are obtained, the computation 
* of integrals of Cn2*h^a over atmosphere (a=0,1,5/3,2) has to be done with 
* at_calcint(). The results are stored internally.

* After the completion of the accumulation time, the integrals of Cn2 by powers 
* of altitude are averaged and converted into seeing, isoplanatic angle and
* effective altitude of turbulence by at_avgint(). Also, the procedure 
* at_calcn2() may be called which uses the average indices and their errors to 
* restore the low-resolution Cn2-profile. 

* Once the profile and atmospheric characteristics are computed, they may be
* saved on disk by at_writeint() and at_writecn2(). The converted average 
* integral values and their errors may be accessed by at_getavgint() and
* at_geter2int() after at_avgint() call.

* Finally, to end with the measurements, call a memory destructor at_done().
*
* @version 1.0: Creation                                                    */
/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _ATMOS_H_
#define _ATMOS_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "weif.h"

/****************************************************************************/
/*                     Declarations of local functions:                     */
/****************************************************************************/

/**
	@brief	Initialize the module local structures for calculations 
	@param	nchan	Number of the MASS detector channels to use
	@param	diam	see wf_setaper()
	@param	eps_inn	see wf_setaper()
	@param	eps_out	see wf_setaper()
	@param	z0  	see wf_setzgrid()
	@param	zmax	see wf_setzgrid()
	@param	dz  	see wf_setzgrid()
	@param	dzmin	see wf_setzgrid()
	@param	responsefile The system transmission function file 
		"lambda[mkm or nm or \AA] P(lambda)"
	@return	0 if Ok or <0 on general error or err=line>0 if problem in \em 
		line-th	row of \em responsefile

	This function should be called, in general, once per the working session, or
	if some basic system parameters (see the function parameters list) are 
	changed. Before the call, at_done() must be invoked.
	
	First, the global structures are initialized, including the normal and
	shifted weight functions sets. Then the weight-related instrumental
	parameters are initialized. Finally, the system transmission function is
	read from the file. The wavelengths, if different from [mkm], are converted
	into micrometers.
	
	@see at_done()
*/
int at_init(
	int nchan, double diam, double *eps_inn, double *eps_out,
	double z0, double zmax, double dz, double dzmin,
	const char * responsefile) ;

/**
	@brief	Load the spectrum from file and check its validity
	@param	sedfile	name of the two-column ASCII file with spectrum 
		"lambda[mkm or nm or \AA] E(lambda)"
	@return	0 if Ok, <0 if general problem (sedfile is not readable or 
		at_init() is not called in advance or spectrum is zero or spectrum does
		not match the MASS response curve with any offset), or err=line>0 if 
		reading	problem occured \em line-th row of spectrum file \em sedfile.
		
	The function calls the local file-reader utility (the same as used for
	MASS  response curve reading). If file is successfully read (seems that
	spectrum does not contain errors as it is), the wavelengths, if different
	from [mkm], are converted into micrometers. Then an attempt is performed to
	multiply this spectrum by response curve read in at_init(). Here the
	negative error code may be returned if no match is found between them. If
	spectrum matches the response curve, it is stored in a global module
	storage to be used in weight functions handling.

	@see at_update()
*/
int at_loadsed(const char * sedfile) ;

/**
	@brief	Check the weight file to fit the currently set system parameters 
	and loaded spectrum
	@param	wfile	weight file name
	@return	0 if file fits, <0 on error
	
	The function wf_leff() is used with current spectrum and then 
	wf_checkfile() is used for the check.
*/
int at_checkwf(const char * wfile) ;

/**
	@brief	Calculate the new weight function set and write it in a file
	@param	wfile	file name for the weight functions set
	@param	progress	progress indicator function (may be NULL in principle)
	@return	0 if Ok, <0 on error
	
	This function first checks the \em wfile with at_checkwf(). If result is
	zero (i.e. "ok"), nothing happens. If not, the weights are calculated with
	currently loaded spectrum and system parameters. Exception is an altitude
	grid parameters. If the checking resulted in error SC_ERZGR (user-supplied
	grid in at_init() does not fit the file grid), weight is calculated with
	user's grid, otherwise - with "optimal" grid. Optimal grid parameters are
	specified in WEIF module.
*/
int at_calcwf(const char * wfile, void(* progress)(int)) ;

/**
	@brief	Update the data structures for atmospheric calculations with a
		new weight function file data
	@param	wfile	The file name of the Weight	Function matrix (see wf_write()) 
	@param	checkfile	file to print the moment approximation information for
		checking (may be NULL)
	@param	zshift1	Altitude shift (in [km], normally \b non-positive) for the
	normal mode or for \em one half of the generalized mode measurements
	@param	zshift2	Altitude shift (in [km], normally \b negative) for the 
		\em another half of the generalized mode measurements
	@return	0 if Ok or <0 on error 

	This function should be called each time the star (i.e. its SED) or the 
	measurement mode (generalized or normal; number of base times per 
	accumulation time), is changed. 
 
	First, if SED is new (last filename is saved internally), it is read from
	the file and cheacked to differ from that for which the WF is currently
	calculated (by wf_issedok()). If SED differs, the weight matrix is tried to
	read from the proposed \em wfile. If nothing good is there and \em
	newwfile==NULL, the function returns AT_ERNOW which means: "no appropriate
	weight file is found on the disk. Check the paths, names etc and start
	at_update() again with newwfile=new_weight_structure_file_name to compute
	the weight from the scratch".

	If weights or \em zshift are changed, then the weights are shifted by \em
	zshift and new approximations for the power-law	dependences of altitude 
	are obtained. The SVD method is used for this. Approximation enable us to
	represent the Cn2-moments by a set of scintillation indices. If \em zshift
	i snon-zero, then both the coefficients for generalized and non-shifted
	modes are obtained. 
	
	A set of test-time definitions is provided in atmos.c to check the
	coefficients of decomposition and the precision of the approximation. The
	coefficients are stored in global arrays to be used in at_calcint().
	
	\note
	The SED function in \em sedfile must be given on a grid with the same step
	as the wavelength grid of the system transmission function (see
	at_init()). Only the start and end-nodes may differ in SED file from
	the response file. All the wavelengths in response file must be present in
	SED file.
	
*/
int at_update(const char * wfile, const char * checkfile, 
	double zshift1, double zshift2) ;
				
/**
	@brief	Allocate the storages for atmospheric integrals calculated by
		at_calcint()
	@param	nmeas	maximal possible number of at_calcint() starts (base times
		per accumulation time)
	@return	0 if Ok, <0 if error (memory alloc. failure)
*/
int at_alloc(int nmeas) ;		

/**
	@brief	Deallocate the memory reserved for module structures
	@return	void
	
	This function releases the memory which is allocated for weights, SED,
	coefficients of Cn2-moments decomposition etc by at_init().
	
	\note
	This function must be kept in strict sync with at_init() and at_update().
*/
void at_done() ;

/****************************************************************************/
/*                    Global atmospheric parameters access                  */
/****************************************************************************/

/** 
	@brief Return calculated instant seeing
	@param	isfree	1: free atmosphere seeing; 0: whole atmosphere seeing
	@return	Seeing [arcsec]
	
	Seeing is computed as 
	\f[
		\beta = AT\_KS \lambda^{-1/5} M_0^{3/5}
	\f]
	where the zero-moment Mo of the turbulence is:
	\f[
		M_0 = \int C^2_n(h) dh
	\f]
	(integration goes from 0 for whole-atm. seeing and starting from \em 
	AT_HBOUND for free-atm. seeing) and \em AT_KS is the calibration constant.
	
	@see at_calcint()
*/
double at_getseeing(int isfree) ;


/** 
	@brief Return calculated instant isoplanatic angle
	@return	Isoplanatic angle [arcsec]
	
	Isoplanatic angle is computed as 
	\f[
		\theta_0 = AT\_KP \lambda^{6/5} M_{5/3}^{-3/5} \sec^{8/5}(z)
	\f]
	where the 5/3-th moment of the turbulence profile is:
	\f[
		M_{3/5} = \int C^2_n(h) h^{5/3} dh
	\f]
	and \f$ AT\_KP=(2.91 \pi^2)^{-3/5} \f$ is the calibration constant.
	
	@see at_calcint()
*/
double at_getisoplan(void) ;
 
/** 
	@brief Return the calculated instant second moment of the turbulence
	@return	Integral of Cn2(h)*h^2 [m^(7/3)]
	
	Second moment of the turbulence is related to the differential astrometry
	precision and the operation of optical interferometers.
	
	@see at_calcint()
*/
double at_getisokin(void) ;
 
/** 
	@brief Return turbulence instant effective altitude
	@param	isfree	1: Free-atm. Heff, 0: whole atm. Heff
	@return Effective altitude [m]
	
	The free atmosphere effective altitude is computed as 
	\f[
		z_{eff} = (M_{\alpha_1} / M_{\alpha_2})^{1/(\alpha_1-\alpha_2)}
	\f]
	where alpha_1 is \em AT_HEFFPOW1 and alpha_2 is \em	AT_HEFFPOW2.
	
	\note
	The value in the whole atmophere is computed, by definition (Heff=M_1/M_0), 
	but nly if shifted indices are available.

	@see at_calcint()
*/
double at_getheff(int isfree) ;

/** 
	@brief Return instant atmospheric time constant
	@return Time constant [ms]
	
	The related to adaptive optics atmospheric time constant is computed from
	Differential Exposure Scintillation Index (DESI) \f$ \sigma^2_{DE}\f$ as
	\f[
		\tau_{de} = AT\_KT \left[ \sigma^2_{DE} \right]^{-3/5}
	\f]
	where \em AT_KT is the empirical calibration constant.
		
	@see at_calcint()
*/
double at_gettimeconst(void) ;

/** 
	@brief Return number of the Turbulence profile altitudes 
	@return Length of the private computed Cn2-array 

	Note that this number is not necessary the length of allocated array
	\em at_cn2[].
	
	@see at_calcn2()
*/
int at_getncn2(void) ;

/** 
	@brief Get pointer to the array of the turbulence profile altitudes 
	@return Pointer to arrays which elements are the altitudes [km] of 
	fitted layers
	
	\note
	Never try to free this array; it is done by at_done().
	
	@see at_calcn2()
*/
double * at_getzcn2(void) ;

/** 
	@brief Get pointer to the array of the turbulence profile layer strengths
	@return Pointer to arrays which elements are the strengths [m^-2/3] indexed
	from 0 to at_getncn2().

	\note
	Never try to free this array; it is done by at_done().
	
	@see at_calcn2()
*/
double * at_getcn2(void) ;

/****************************************************************************/
/*                     Calculation of atmosphere parameters                 */
/****************************************************************************/

/**
	@brief Compute the atmosphere integral parameters and Cn2-profile
	@param	scind	array of non-shifted indices
	@param	scinds	array of shifted indices (may be NULL)
	@param	meas	current measurement number (base-time within 
	current accumulation time)
	@return	0 if Ok, <0 if error
	
	Using the Cn2-moments decomposition coefficients and  scintillation
	indices, the calculation of integral atmospheric parameters is performed.
	Number of indices in either \em scind or \em scinds must be equal to number
	of respective computed weights in \em at_wf which are made in at_init()
	call. No check of the validity of supplied \em scind(s) arrays is done.
	
	The coefficients of multiplication of the indices in normal and shifted
	mode are already precomputed in at_init(); the results become accessible
	after the function call by at_getseeing(), at_getisoplan(), at_getisokin(),
	at_gettimeconst() and at_getheff(). The atmospheric time constant is
	computed given the \em desi parameter.  If \em scind(s) or \em desi
	parameters are empty (NULL or 0, respectively), the calculations for
	supplied not-NULL parameters are only made.

	The computed quantities are saved in the \em meas-th line of the global 
	storage for further averaging	by at_avgint(). Normally, \em meas has to be
	substituded with the function sc_getcurmeas() result in the call to 
	at_calcint().
	
	\b Performance
	
	One calculation of integrals takes \b 5 ms at PC P-III 667 Mhz.

*/
int at_calcint(double * scind, double * scinds, int meas) ;

/**
	@brief Restore the low-resolution Cn2 profile
	@param	scind	array of average non-shifted indices
	@param	scinds	array of average shifted indices (may be NULL)
	@param	e2scind	array of \em squared errors of \em scind
	@param	e2scinds	array of \em squared errors of \em scinds
	@param	isfixed	Use fixed-altitude layers restoration rather than strongest
	layers search
	@return Minimal chi-square reached, or (double)(error_code,<0) on error
	
	The profile restoration results in altitudes of the at_getncn2() layers 
	altitudes at_getzcn2() with strengths at_getcn2(). Two methods are
	implemented - the search for four strongest layers (if not \em isfixed) and
	the search for the intensity of six layers placed proportionally in the
	altitude range with relative resolution 0.5.

	\b Performance
	
	One restoration takes typically \b 45 \em ms at PC P-III 667 Mhz for the
	method with 6 \em fixed layers, and \b 140 \em ms for placing 4 \em 
	floating altitude layers.

*/
double at_calcn2(double * scind, double * scinds,
	double * e2scind, double * e2scinds, int isfixed) ;
	
/**
	@brief	Average integrals, compute their standard errors and write to 
	disk
	@param	nmeas	Actual number of measurements done (e.g. sc_getcurmeas())
	@param	desi	average DESI for shifted channel A
	@param	e2desi	squared error of \em desi
	@param	zdist	zenith distance in [degree]
	@return 0 if Ok, <0 on error
	
	Given the instantaneous integrals in their storage, their average values and
	standard deviations are computed and put in respetive global vectors.
	The function sc_avgmatrix() is used for averaging. Errors thus take into
	account the correlation of values.

	Then the average integrals are converted into atmospheric parameters -
	seeing, isoplanatic angle and effective altitude of turbulence. They are
	available via at_getseeing(), at_getisoplan(), at_getheff() and
	at_gettimeconst() functions. Also, the atmospheric time constant is computed 
	from provided \em desi and \em e2desi values.

*/
int at_avgint(int nmeas, double desi, double e2desi, double zdist) ;

/**
	@brief	Write average atmospheric parameters, the atmospheric time constant
	with their errors and the restored Cn2 profile on disk
	@param	f	FILE pointer
	@param	stamp	leading word in the output line
	@return	0 if Ok, <0 on error
	
	The single line is written on disk, consisting of average atmospheric
	parameters each followed by its error. Parameters are free seeing ["],
	seeing ["],	isoplanatic angle ["], second moment of turbulence [m^{4/3}], 
	and atmospheric time constant [ms].
*/
int at_writeint(FILE * f, const char * stamp) ;

/**
	@brief	Write the restored Cn2 profile on disk
	@param	f	FILE pointer
	@param	stamp	leading word in the output line
	@return	0 if Ok, <0 on error
	
	The Cn2-profile data line consist of 1) the number of adjusted Cn2 layers, 
	2) Chi-square 3) the at_getncn2() altitudes of layers and 4) their
	at_getncn2() strengths in [m^-2/3].
	
*/
int at_writecn2(FILE * f, const char * stamp) ;

/**
	@brief	Get the pointer to the array of average integral values
	@return	array of averages

	Use the \em at_ordint enumeration members to access the proper element 
	in the returned array.
*/
double * at_getavgint() ;

/**
	@brief	Get the pointer to the array of \em squared errors of average 
		integral values
	@return	array of error^2 of averages

	Use the \em at_ordint enumeration members to access the proper element 
	in the returned array.
*/
double * at_geter2int() ;

/*==========================================================================*/
/*                             Definitions                                  */
/*==========================================================================*/

/** Number of modes: normal and generalized, hense = 2 */
#define AT_NMODE	2

/** Minimal ratio of the singular value to the maximal one */
#define AT_MINRELW	1e-4

/** Maximal number of orthonormal functions to use in h-powers restoration: (to
apply after AT_MINRELW based rejection) */
#define AT_MAXNBASE	10

/** Binary mask for using the shifted weights: "1"="use it".
	 Now we thus use all non-shifted and As, Bs and ABs, ACs weights. */
#define AT_USESHFT "1100100000"
//#define AT_USESHFT "1111111111"

#define AT_POWSEE	0		/**< Power of Cn2 moment for seeing */
#define AT_POWISP (5./3.)	/**< Power of Cn2 moment for isoplanatic angle  */
#define AT_POWEFF (1.0)		/**< Power of Cn2 moment for Effective altitude */
#define AT_POWISK (2.0)		/**< Cn2 second moment power */

/** Height of the boundary layer [kilometers] to fit the Cn2-integral for 
free seeing (h^0 in integral is forced to resemble the AB-pair weight function
below this height) */
#define AT_HBOUND (1.0)

/** Number of power laws to fit: 2 kinds of 0-th, 5/3-th, 1-st, 2-nd */
#define AT_NPOWER 5

/** Arrangement of the moment decomposition coefficients in the at_coef[] array: 
it's essential for at_calcint() that AT_IWSEE goes in advance to AT_IFSEE. */
typedef enum {
AT_IFSE,	/**< at_coef[Free seeing][]: */
AT_IWSE,	/**< at_coef[Seeing][] */
AT_IISP,	/**< at_coef[Isoplanatic angle][] */
AT_IEFF,	/**< at_coef[Turbulence effective altitude][] */
AT_IISK 	/**< at_coef[Isokinetic angle][] */
} at_ipow ;

/** Calibration constant in seeing formula (AstL 45, p.399): 
\f$ seeing[''] = AT\_KS * \lambda_{eff}^{-1/5} M_0^{3/5} * 
\sec^{-1}(z)\f$, where \em lambda_eff is in [mkm]*/
#define AT_KS 1.73588e+07

/** Calibration constant in isoplanatic angle formula (AstL 45, p.399)
\f$ \theta_0[''] = AT\_KP * (M_2 /\lambda_{eff}^2)^{-3/5} * \sec^{8/5}(z) \f$, where
\f$ AT\_KP = (2.905*(2\pi)^2*(MKM2M)^{-2})^{-3/5}*206265 \f$*/
#define AT_KP 0.000756348

/** Atmospheric time constant empirical calibration constant for DESI for 1,3ms
integrations in diameter=2cm apertures with -1km shift */
#define AT_KT 0.175

/** Cn2-profile restoration weight matrix spacing: 1km (integer-like!)*/
#define AT_DZCN2 1

/** Order of atmosphere integrals in their storage matrix and output file line.
Second line contains the errors of respective values. AT_LAST must stay always 
at the end*/
typedef enum {AT_FSEE,AT_WSEE,AT_ISP,AT_M2,AT_FHEFF,AT_HEFF,AT_TC,AT_LAST}
at_ordint ;

/** Formats of the records of average integral and their errors in the file. 
The order of strings follows the order of integrals specified in \em at_ordint.
Each format must be preceded with "\%s" to insert a delimiter specified in 
AT_INTDLM.
*/
#define AT_INTFMT "%s%6.3f","%s%6.3f","%s%6.3f","%s%8.2e","%s%5.0f","%s%5.0f","%s%5.2f"

/** Delimiter of entries in the average data file */
#define AT_INTDLM	" "

/** Altitude of layers record format with leading "\%s" for a delimiter */
#define AT_ALTFMT "%s%2.0f"

/** Profile values format with leading "\%s" for a delimiter */
#define AT_CN2FMT	"%s%8.2e"

#endif
