/*--------------------------------------------------------------------------*/
/** @file scind.h
*                  MASS project: TURBINA module file                       
*                       header file for scind.c                             
*                                                                          
* The module SCIND is aimed to implement the calculation of stellar 
* scintillation indices from individual counts of several photo-multiplier 
* tube (PMT). It takes into account the following supplementary information 
* apart from the count arrays:
* - the value of sky \em background level in pulses/ms;
* - the \em dead-time of the PMT to account for non-linearity;
* - the \em non-Poisson factor (close to 1) which converts the signal
* mean value into dispersion (which is an exact unity for an ideal detector).
*
* As a result, the \em normal scintillation index is produced which is free 
* from photonic statistics influences and detector imperfections. The 
* differential indices which deal with signals \em covariation from different
* channels are also  computed. The notions of "aperture" and "channel" are 
* identical here.
*
* This code originates from the program \b select.c written by V.Kornilov for
* DASS project evaluation. All the functions and global variables dealing with 
* scintillation indices have a prefix \em sc_. 
*
* \b Usage:
*

 For the sake of performance and simplicity, the work-arrays which are needed
 in calculations of indices are allocated only once, by the initiation utility
 sc_init(). Then all the needed computations are done by sc_compute() a certain
 number of times, during which the indices should be saved to disk by 
 sc_writeidx(). If the value of some parameter of sc_init() has changed, this
 function may be called again, with no call to sc_done() in advance. The needed
 space will be reallocated if needed. In general, sc_init() may be called whenever
 it's convenient, without loss of time of memory. 
 
 Indices, computed by sc_compute() (i.e. by sc_indices()) may be accessed by
 sc_getidx(). 

 After some accumulation of indices (in local index storages), the indices can
 be averaged by sc_average(). These average indices, corrected to zero
 exposure, can already be used for calculations of atmosphere models and
 parameters by external means (see at_calcint() in the module "atmos"). Also,
 they may be saved to disk with sc_writeavgidx(). After using these indices,
 the counter of accumulated indices must be reset by sc_again() before begin of
 the next accumulation time and then the cycle of instantaneous index
 computations with sc_compute() can be restarted. After finishing the job, the
 memory should be cleaned with sc_done().

\note
 In principle, index averaging may be done in any moment providing thus some
 on-fly smoothing of data in a sliding window. It is possible since the index
 storages work as some cycle buffers, i.e. the rows which are not covered in a
 current assumulation time still contain information from previous accumulation
 time. Empty rows of index storages won't be used, since the measurement
 mode-keeping storage of the module is reset to "mode=unknown" each time
 sc_init() is called.

*
* @author N. Shatsky, Sternberg Institute (kolja@sai.msu.ru)               
*
* @version 1.0: Creation from select.c                                      */
/* This header file is equipped with documentation made in comments in a
* format suitable for production of LaTeX or HTML documentation with the
* doxygen software.
* Editor tabular size: 4                                                    */
/****************************************************************************/

#ifndef _SCIND_H_
#define _SCIND_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "nr.h"

/** Type of PM tube counts  */
typedef int count_t;

/** This structure contains the PMT count series for which the scintillation
index should be computed. To simplify the data handling, the descriptive
parameters of this series and the detector are included in this structure.
The length of the data buffer is not specified since it may vary. */
typedef struct _SC_CHAN_T_ {
	count_t * data ;		/**< Array of counts */
	double nonpois ;		/**< non-Poisson factor (dispersion/mean) */
	/** background level (scaled to the used micro-exposure time) */
	double bkgr ; 			
	double deadtime ;		/**< deadtime (divided by micro-exposure) */
} sc_chan_t ;

/** Enumeration of entities to access or write on the disk, to use in
sc_get...() and sc_write...() functions. */
typedef enum SC_WHAT {
	SC_HEAD,	/**< Head line writing */ 
	SC_DSI, 	/**< Aperture indices writing (with DESI) or access */
	SC_DESI, 	/**< Differential exposure indices access */
	SC_AFLUX,	/**< Average flux in channels access or writing */
	SC_MOM  	/**< Count moments writing */
	} sc_what ;
/** Formula to get a Number of entrance apertures and their non-redundant 
combinations having \em a apertures */
#define NAPCOMB(a) (((a)*(a)+(a))/2)

/** Use true logarithmic calculations of scint. indices instead of relative
differential calculations */
//#define SC_LOGIDX


/****************************************************************************/
/*                     Declarations of local functions:                     */
/****************************************************************************/

/**
	@brief	Gives a name for a given index of entrance aperture or 
	combination of apertures
	@param	nchan	number of apertures
	@param  index	sequential number of aperture
	@return	pointer to name if Ok, or NULL if invalid index or NULL wf
	    
	Name of an aperture corresponds to the given index according to
	convention: For index=0..nchan-1: name='A','B',... (for normal indices
	of scintillation in a single aperture); for index=nchan: name='AB'
	(first combination of apertures), nchan+1: 'AC',...,'BC','BD' and so
	on.

	This name is copied in the static character string which contains
	characters (two for the combination and one for trailing zero) and
	returns the pointer to this string. Thus, if more than one call to this
	function is done in one statement, the result is unpredictable.

*/
char * sc_apername(int nchan, int index);

/**	
	@brief	Computes the normal and differential scintillation indices
	@param	sets	array of sc_chan_t structures for a set of channels
	@param	nsets	number of elements in \em sets (channels)
	@param	ndata	number of counts put in channel buffers for processing
	@param	ncorr	array of numbers of corrected counts in channels 
		(replaced with mean if missing) 
	@param	scind	array of resulting scintillation indices (NAPCOMB(nsets)
	elements.)
	@param	desi	array of resulting differential exposure indices
	@return	0 if Ok or <0 if error

	This utility computes a set of all \em instantaneous (i.e. determined using
	individual PMT counts obtained during a \em base \em time) scintillation
	indices (SI) which can be derived for a given number of channels. Thus, 
	both \em normal and \em differential indices (DSI) are calculated.
	
	The implementation of this function depends on the \em SC_LOGIDX 
	definition: if it's set, then the "true" calculations of variations of 
	logarithms of counts is implemented; otherwise - the relative differential 
	approximation is used. Here below we refer to these alternative modes of 
	implementation as "\em if \em SC_LOGIDX".
	
	Calculations of scintillation indices go through following steps:
	- compute the ``raw'' means of counts 
	- if \em SC_LOGIDX, compute logarithms of all individual counts and their 
	means;
	- calculate dispersions and covariations of counts or (if \em SC_LOGIDX) of 
	their logarithms;
	- calculate also \em sc_r1 and \em sc_r2 - cross- and 
	auto-covariations of signals with the lag of 1 and 2, respectively.
	- if \em SC_LOGIDX, denormalize the dispersions by the squared mean count and
	covariations by the product of mean counts in two channels ;
	- correct them and mean counts for non-linearity (dead-time of PMT)
	- compute scintillation indices free from photon noise (using non-Poisson
	factors for conversion of means into photon noise dispersion)
	- these indices are computed corrected to zero-exposure
	(as \em 1.5*r0-0.5*r1, where r0 and r1 are dispersions (for normal indices)
	or covariations (for differential indices) with	lags 0 and 1, respectively.
	Also, DESIs are computed for all channels, using auto-covariations with lags
	0 (i.e. dispersion), 1 and 2.
	
	See ``Multi-Aperture Scintillation Sensor. Final design report'', Appx. C
	(p.87).
	
	The counts from channels sometime miss a certain a number of data due to the
	data loss in the line. In this case, the "data provider" is supposed to
	replace them with the average count in respective channel. Thus, \em ncorr[]
	is used to only correct the dispersion of counts, by decrement of the number
	of data in calculations.
		
	The order of indices in which calculated indices are assigned to elements
	of an output array \em scind is exactly the same as order of going of \em
	weights in weight matrix which is computed in module weif.c : first
	normal indices then the covariation of first channel data with second, with
	third etc..., of second channel with third etc. See sc_apername() for exact
	sequence.

	\b Performance:
	
	The test consists of calculation of scintillation indices for 4 channels
	including two DESI calculations for channels of innermost entrance
	apertures. It is actually a call to this function  by sc_compute(). 
	It takes:

	- In relative approximation of indices (normalized dispersion or 
		covariation):	
		- \em 1.7ms at P-III 667 MHz,
	- In true logarithmic form (\em SC_LOGIDX definition set):
		- \em 3ms at P-III 667 MHz,
	
	
	\note 
	- the resulting indices in \em scind are \b already \b 
	corrected \b to \b zero-exposure;
	- The code assumes the background level and dead-time to be supplied 
	in the same	time-scale as the series counts; if binned data are supplied,
	the respective deadtime should be given divided by bin-size;
	- Program currently refuses to develop the sets which have different 
	length;
	- Procedure uses pre-allocated arrays \em sc_mean, \em sc_cmean, \em 
	sc_sig to store the results of statistical calculations. They contain,
	respectively: mean counts and means corrected for non-linearity,
	dispersions and covariations of counts corrected for non-linearity. Note,
	that, the \em i-th element of the dispersion/covariation array addresses
	the same channel or combination of channels as \em i-th element of the
	output scintillation index array. The public arrays \em	sc_r1 and 
	\em sc_r2 are also used and are accessible similarly to \em sc_sig.
	These arrays may be accessed immediately after processing.
	- The array \em scind must contain AT LEAST \f$(nsets^2+nsets)/2\f$
	elements. An array \em desi must contain \em Nchannel elements. 
	Otherwise, program crushes. 
	
	\todo
	Allow the processing of channel data series which have different 
	micro-exposure and, hence, length. It can be needed to increase the time
	resolution in measurements with channels of the small entrance aperture.
	
	@see sc_init() sc_apername() sc_compute()
*/
int sc_indices(sc_chan_t * sets, int nsets, int ndata, int * ncorr,
	double * scind, double * desi) ;

/**
	@brief	Compute the theoretical value of the normal scintillation index for
	MASS statistical test evaluation
	@param	data	Array of modulated channel counts
	@param	n	length of \em data
	@param	scind	Pointer to the resulting index
	@return 0 is Ok, <0 on error
	
	MASS statistical test is performed using the specially modulated flux from
	the control LED. The current through the control LED is modulated step-like
	sinchronously with counting time diagram with a period of four
	microexposure. During each first and third takts of this modulation, the
	current is at the middle level, whilst at second and fourth it's increased
	and decreased, respectively, relative to this middle level. 
	
	The task of this utility is to determine these middle, increased and
	decreased light levels and to compute the ("theoretical") scintillation index
	which corresponds to this modulation (see sc_indices()):
	
	\f[
		s^2_{th} = \frac{1.5 \sigma^2 - 0.5 \rho_1}{\overline{flux}^2}
	\f]
	
	Here, for the specified way of modulation, the dispersion is calculated as
	\f[
		\sigma^2 = \frac{2\Delta_0^2 + \Delta_+^2 + \Delta_-^2}{4}
	\f]
	and the autocovariation with the unity lag is
	\f[
		\rho_1 = \frac{\Delta_0 (\Delta_+ + \Delta_-)}{2}
	\f]
	and Deltas with subscripts "0", "+" and "-" are the deviations of the
	middle, increased and decreased light levels, respectively, from the
	average level.  Middle level may differ from average level if the positive
	and  the negative variations are different in magnitude.
	
	\note
	Since the flux modulation is the same in all the channels of MASS,
	cross-channel differential indices measured during the test (by
	sc_indices()) must be about zero within their errors.

*/
int sc_stattest(count_t * data, int n, double * scind) ;

/**
	@brief Recurrence computer of an average and dispersion of the count series
	@param	data	Current array of channel counts
	@param	n	length of \em data
	@param	k	sequential number of the array \em data in a measurement session
	@param	avg	Pointer to running average count,	[count]
	@param	sig	Pointer to running dispersion of counts, [count^2]
	@param	isfilt	To account for the long-term modulation of the signal
		(clouds, etc.) in dispersion \em sig
	@return 0 is Ok, <0 on error

	This utility may be used either as a plain average and dispersion computer
	and as a tool for calculation of the statictics of a long series consisting
	of multiple data pieces (arrays \em data). In both instances, the first call
	to the utility must have \em k=1. If so, the values \em avg and \em sig are
	assumed to contain no information about the previous data from random
	process (arrays \em data). Then they are assigned like follows:
	\f[
		avg = \overline{data} = \frac{1}{n}\sum\limits_{i=0}^{n-1} data[i] ,
	\f]
	\f[
		sig = \sigma_{data} = \frac{1}{n-1}\sum\limits_{i=0}^{n-1}
		(data[i]-\overline{data})^2
	\f]

	When one wishes to obtain the statistics of a continued series of arrays
	\em data (which, obviously, represent the realisation of the same random
	process), the values \em avg and \em sig are updated for each consequent \em
	k-th array \em data (\em k>1) like follows:
	
	\f[
		avg_{new} = \frac{1}{k} \left( (k-1) avg + \overline{data} \right) ,
	\f]
	\f[
		sig_{new} = \frac{1}{k} \left( (k-1) sig + \sigma_{data} \right) + Corr ,
	\f]
	where the correction \f$ Corr=(k-1)*(avg_{new}-avg)^2 \f$ of the dispersion
	for the low-frequency  modulation of the average count is computed and
	added if \em isfilt==0. Note, that \em k must be incremented by user.

	An estimate of the non-poissonity \em P of the detector and an error of the
	running average \em avg may be obtained manually as:
	\f[
		P = \frac{sig}{avg} ~,~~
		\sigma_{avg} = \sqrt{\frac{sig}{k  n}} ,
	\f]
	provided that \em k arrays \em data are treated by this utility and they all
	have the same length \em n (if not, the last denominator is replaced by
	the sum of all the \em n-s of the used arrays).
*/
int sc_calcstat(count_t * data, int n, int k, double * avg, double * sig, 
	int isfilt) ;

/**
	@brief Average a matrix of values with rows selection along second dimention
	@param	matrix	matrix[nrow][ncol] of values to average along second 
		dimention (i.e. to obtain an average row). Meanwhile, if ncol==0, 
		\em matrix is interpreted as a vector to be averaged along itself.
	@param	ncol	number of columns in \em scind_set[][]
	@param	nrow	number of rows in \em scind_set[][]
	@param	rowselect	row selection vector[\em nrow] (set NULL to select all)
	@param	select	selection value in \em rowselect (no matters if
		\em rowselect==NULL )
	@param	avg 	vector[\em ncol] of resulting average values
	@param	err2	vector[\em ncol] of resulting squared errors of avg.values
	@param	lag  	maximal lag in correction of error for the correlation of
		values
	
	The rows are selected for averaging from the storage \em matrix by
	comparison of respective values in \em rowselect with \em select up in a
	range of the storage row numbers [0..nrow-1]. Average value is thus derived
	for each column for rows with index \em row	complying the \em
	rowselev[row]==select condition. 
	
	Errors of average values are computed corrected for correlation of values
	as given in Korn & Korn "Spravochnik po matematike" 4th edition, section
	19.8 with lag of autocovariation up to \em lag. If a given \em lag is more
	than \em nrow, the maximal lag is taken as \em nrow-1.
	
	@see sc_average
*/
int sc_avgmatrix(double ** matrix, int ncol, int nrow, 
	signed char * rowselect, signed char select, 
	double * avg, double * err2, int lag) ;

/**
	@brief	Average scintillation indices and compute their \em squared errors
	@param	isgen	0: no generalized mode was used, 1: with generalized mode
	@param	lag  	Maximal counts lag to account for the index correlation in
	the error of mean (SC_MAXLAG may be set)
	@return	0 if Ok, <0 if bad parameters or NULL arrays
	
	This utility is called normally after \em accumulation \em time span during
	which rows of the index matrices are computed sc_getcurmeas() times times.
	It is aimed to obtain some \em mean value of each index or, in
	other words, to implement the averaging of \em scind_set  matrix along its
	time-dimension. Also, the squares of uncertainties of these indices are 
	assessed.
	
	The computed average indices, average shifted indices, and DESIs (both
	normal and shifted) are put in the global vectors, to be accessed by
	sc_getavgidx() and sc_geter2idx().

	The error of mean for the index \em x is computed as follows:
	\f[
	\sigma^2_{\bar x} = \frac{1}{N} \sigma^2_x 
	\left[1+2\sum\limits_{k=1}^{lag} \rho_k(x)\right]
	\f]
	where the \f$\rho_k\f$ is the covariation of indices \em x with the 
	lag \em k:
	\f[
		\rho_k(x) = \frac{1}{N-k}\sum\limits_{i=1}^{N-k}x_i x_{i+k}
	\f]
	
	\note 
	- To the moment, the average index itself is a simple mean with no 
	rejection; it can in principle be implemented as a mean with rejection 
	or as a median.

*/
int sc_average(signed char isgen, int lag) ;

/**
	@brief	Initialize the structures to put the scintillation indices
	@param	maxndata	Maximal expected number of counts in base time 
	series (one simultaneous measurement of \em nchan channels)
	@param	maxnmeas	Maximal expected number of measurements 
	(accum.time/base.time) to keep in the storage
	@param	nchan	Number of channels (entrance apertures)
	@return	0 if Ok or <0 on error
	
	The module contains the arrays \em sc_dsidx and \em sc_desidx which keep,
	respectively, the normal and differential indices and the
	differential exposure indices. These are the matrices in  which the rows
	contain all simultaneously obtained indices for one particular base time.

	For the sake of a simple code structure, some intermediate data processing
	is done in advance to calculation of indices, i.e. the data binning,
	or smoothing is done separately and before the calculations of indices. 
	For this and for acceleration of logarithmic indices derivation, some work 
	arrays are also allocated.
	
	DESI are computed for all apertures.
	
	The dimentions of allocated arrays are determined by the numbers \em
	maxndata, \em maxnmeas and \em nchan which are copied in
	private module global variables. Another private global variable - \em
	sc_curmeas - counts the performed index calculations and initialized to
	zero here. The number of counts in any subsequent base-time series should
	not exceed the value of the parameter \em maxndata; ditto for the
	measurements number (base times per accumulation time) and the parameter \em
	maxnmeas.

	This function should be called, obviously, before the first calculation of
	indices once the number of expected series (base-time measurements within
	one accumulation time) is known (configuration file is read). In addition to
	sc_init() call, ALL the members of (private to module) channel parameter 
	structures should be initialized by \em nchan calls to sc_chan_init().
	
	If the parameters specify the storage dimentions which are less than that
	which were previously allocated, the storages are left unchanged; if less or
	storages are not allocated at all, the allocation takes place.

	\note
	This function must be kept in strict sync with sc_compute().

	@see sc_chan_init() sc_compute() sc_done()
*/
int sc_init(int maxndata, int maxnmeas, int nchan) ;

/**
	@brief 	Initialize channel data in static channel array element
	@param	chan	sequential number of the channel (0,1..Nchannel-1)
	@param	data	pointer to the buffer - receiver of channel PMT counts
	@param	nonpois	non-poisson factor (near unity)
	@param	bkgr	background level in [counts per microexposure time]
	@param	deadtime	deadtime of PMT \em divided by microexposure time
	@return 0 if Ok, <0 if error
	
	Function initializes the \em chan-th element of the pre-allocated array \em
	sc_channel of  the channel structures type \em sc_chan_t. This array is
	expected to be already allocated by the call to sc_init().

	The assignement is made only if the respective parameter is different from
	null or zero. Thus, a single all-initializing call may be split into a
	sequence of calles which initialize (or update) some particular members 
	(e.g., changing the background level in a channel).
	
	\note
	No memory allocation is done here for data[] array, it is only pointed to 
	already	allocated buffer for receiving the channel counts.
	
	@see sc_init()
*/
int sc_chan_init(int chan, count_t * data, double nonpois, double
bkgr, double deadtime) ;

/**
	@brief	Make all the calculations to compute the indices of interest
	@param	ndata	Actual number of counts in the channel buffers
	@param	ncorr	array of numbers of corrected counts in channels 
		(replaced with mean if missing) 
	@param	isgen	Generalized mode flag
	@return	0 if Ok, <0 if called functions return error or sc.indices arrays
	overfull
	
	A single call to sc_indices() is made which performes all the needed 
	calculations. For performance issue, see sc_indices() documentation. The
	results of index calculations are put in the current measurement's row (see
	sc_getcurmeas()) of the local index storing matrices.
	
	The parameter \em ndata must not exceed the value \em maxndata with which
	the sc_init() was lastly called; \em ncorr[chan] collects the number of 
	counts in respective channels \em chan replaced with mean value due to their
	loss in the line.

	The calculations are made irrespective of \em isgen parameter of this
	function. This parameter is only saved in additional local storage to be
	able to disentangle the normal and generalized data while averaging the
	indices.
	
	Before exiting and upon the success of calculations the counter \em
	sc_curmeas of performed measurements is incremented.  If the value of \em
	sc_curmeas equals or exceeds the declared in sc_init() maximal number \em
	maxnmeas (the length of allocated storages for indices), then an error is
	returned.

*/
int sc_compute(int ndata, int * ncorr, signed char isgen) ;

/**
	@brief	Get the number of measurements done in a current accumulation time
	@return	measurement number or <0 on error
	
	Example: if sc_again() was just used, the function will return "0" since no
	measurement were done up to now.
*/
int sc_getcurmeas() ;

/**
	@brief	Reset the counter of accumulated indices in local index storages
	@return	void
	
	Once the indices in matrices sc_dsidx[][] and sc_desidx[][] are not needed
	any more, the counter of current row in them should be reset to avoid the
	matrices overfull. Do this once the indices are averaged, saved and 
	displayed.

	@see sc_writeidx() sc_average()
*/
void sc_again() ;

/**
	@brief	Deallocate the memory reserved for module index storages
	@return	void
	
	This function releases the memory which is allocated for index storages by
	sc_init() according to the module variables \em sc_nmeas and \em sc_nchan.
	
	This function should be called only to end with calculations at all, since
	the modifications of arrays may be done by the calls to sc_init() with no
	intermediate sc_done().

	\note
	This function must be kept in strict sync with sc_init().
*/
void sc_done() ;

/**
	@brief	Write the latest instantaneous (base-time related) indices onto the 
	disk
	@param	f    	Disk file structure pointer
	@param	what	SC_DSI: write the last obtained index set (both (D)SI and 
		DESI), SC_HEAD: write the header line
	@param	stamp	Reference string of Mode/UT to put in the first column
		(if \em what!=SC_HEAD) or its column name (if \em what==SC_HEAD)
	@return 0 if Ok, <0 if error
	
	Function writes the last computed indices from matrices \em sc_dsidx[][]
	and \em sc_desidx[][] on the disk in one ASCII line. The first token in the
	line is the reference  \em stamp (which, e.g., contains the comment
	symbol "# ",  Normal/Generalized mode symbol "N/G " and some word with the
	UT hour).

	If \em what is SC_HEAD, the header is created like follows:
	\verbatim
<stamp (name)> SI(A) SI(B) .. DSI(AB) ....... DESI3(A) ..
	\endverbatim
	where SI relate to normal indices, DSI - to differential indices, DESI3 are
	differential exposure indices with smoothing 3-element window. Header line 
	is kept in sync with what is written in data-writing mode (\em 
	what!=SC_HEAD).

	For writing the index values, the format SC_IDXFMT is used. Numbers are 
	delimited with SC_IDXDLM. The header-writing call is only possible after 
	the sc_init() was called; data-writing is allowed after at least one
	sc_compute() call. 
	
	The index set is written in one line, which has a structure corresponding
	to that given in the header: \em stamp, \em Nchannel normal indices, 
	\em	NAPCOMB(Nchannel)-Nchannel differential indices, and 
	finally the differential exposure indices (DESI) for all channels.

	\note
	- This function must be kept in strict sync with sc_compute().
	- The sc_writeidx() must not be called if sc_again() was already called 
	after the sc_compute() call.

	@see sc_init() sc_compute() sc_again()
*/
int sc_writeidx(FILE * f, sc_what what, const char * stamp) ;

/**
	@brief	Write the \em average (accumulation-time related) indices on the 
		disk
	@param	what	SC_DSI:  write the average indices, SC_AFLUX: write the 
		average fluxes in channels
	@param	isgen	0: write non-shifted values with errors, 1: write shifted
		values with errors
	@param	f	Disk file structure pointer
	@param	stamp	Reference string to put in the first column
	@return 0 if Ok, <0 if error
	
	Function writes the last computed average indices from vectors \em
	sc_adsi[] \em sc_adesi[] and their errors on the disk in one ASCII line.
	The first token in the line is the reference  \em stamp, which normally
	contains the comment symbol "# ",  and whatsoever.

	For writing the index values and their errors, the format SC_IDXFMT is 
	used. Numbers are delimited with SC_IDXDLM. 
	
	The index set is written in one line, which has a following structure: \em
	stamp, average normal and differential indices for all channels and their
	combinations, each followed by its error and  average differential exposure
	indices (DESI) each followed by its error. The sets are taken from shifted
	or non-shifted parts of storages, according to the parameter \em isgen.

	\note
	- This function must be kept in strict sync with sc_average().

	@see sc_init() 
*/
int sc_writeavgidx(FILE * f, sc_what what, signed char isgen, const char * stamp) ;

/**
	@brief	Write the (base-time related) signal moments on the disk
	@param	f  	 	Disk file structure pointer
	@param	what	SC_MOM: write the moments, SC_HEAD: the header line only
	@param	stamp	Reference string of Mode/UT to put in the first column
	(if \em what!=SC_HEAD) or its column name (if \em what==SC_HEAD)
	@return 0 if Ok, <0 if error
	
	Function writes the signal moments - means and dispersions and
	covariations on the disk in one ASCII line. The structure of the code is
	very similar to sc_writeidx(). The first token in the line is the
	reference  \em stamp (which normally contains the comment symbol "# ", 
	Normal/Generalized mode symbol "N/G " and some word with the UT hour.

	If \em what is SC_HEAD, the header is created like follows:
	\verbatim
<stamp (name)> MEAN(A) MEAN(B)... SIG(A) SIG(B) .. COV(AB) ... COV1(A) ... COV2(A) ..
	\endverbatim
	where SIG relate to dispersions and COV to covariations of channels and
	to auto- and cross-covariation of the signal in channels with lag 1 and	2. 
	
	\note
	
	For writing the moment values, the format SC_MOMFMT is used. Numbers are 
	delimited with SC_MOMDLM. The header-writing call is only possible after 
	the sc_init() was called; data-writing is allowed after at least one
	sc_compute() call. 
	
	The moments' set is written in one line, which has a structure
	corresponding to that given in the header: \em stamp, means for all
	channels, \em NAPCOMB(Nchan)-Nchan normal dispersions and covariation, then
	the same number of lag-1 auto- and cross-channel covariations, and finally
	\em Nchan auto-covariations with lag 2.

	\note
	- This function must be kept in strict sync with sc_compute().
	- The sc_writemom() can be called only immediately after sc_compute(), since
	the	moments are stored for current calculations only.

	@see sc_writeidx()
*/
int sc_writemom(FILE * f, sc_what what, const char * stamp) ;

/**
	@brief Scind and Weif modules error messages
	@param	error Error code returned
	@return pointer to static character string with error message or NULL
	    
	Use as sc_error(what_was_returned_if_negative). Returns NULL in case
	of unknown/unused error code.
*/
char * sc_error(int ercode) ;

/** 
	@brief	Get pointer to the index array
	@param	meas	Measurement (row number of the index matrix)
	@param	what	0 for usual index, 1 for DESI
	@return	pointer to 1st element of the index array or NULL on error
	
	One may use sc_getcurmeas()-1 for the parameter \em meas, to return the
	last obtained measurement result. Indices are ordered in the returned
	vector as specified in sc_apername() documentation; the length of vector is
	NAPCOMB(Nchannel) or \em Nchannel for  what==1 (i.e. DESI).

	\note	Never try to free the returned pointer to array. It is done by
	sc_done().
	
*/
double * sc_getidx(int meas, sc_what what) ;

/** 
	@brief	Get pointer to the averaged index array
	@param	what	SC_DSI: for usual index, SC_DESI: for DESI
	@param	isgen	0 for array of non-shifted average indices; 1 for shifted
	indices
	@return	pointer to 1st element of the index array or NULL on error
	
	The call with \em isgen==1 is only possible if generalized mode was used
	during last accumulation time.
	
	@see sc_getidx()
	
*/
double * sc_getavgidx(sc_what what, signed char isgen) ;

/** 
	@brief	Get pointer to the array of the \em squared errors of average index
	@param	what	SC_DSI: for usual index, SC_DESI: for DESI
	@param	isgen	0 for array of non-shifted average indices; 1 for shifted
	indices
	@return	pointer to 1st element of the index array or NULL on error
	
	@see sc_getavgidx()	
*/
double * sc_geter2idx(sc_what what, signed char isgen) ;

/** 
	@brief 	Get pointer to the array of mean counts in channels in [counts per
	basetime]
	@param	iscor	0: return raw mean; 1: return non-linearity corrected mean
	@return	Array of mean counts in channel [counts per microexposure]

	Valid indices (channel numbers) of the returned array are [0..Nchannel-1].
	See noted to sc_getidx().
	
*/
double * sc_getmean(int iscor) ;

/** 
	@brief 	Get pointer to the array of average mean corrected counts in 
		channels in [counts per basetime]
	@param	isgen	0: non-shifted average counts; 1: shifted averages
	@return	Array of average mean counts in channel [counts per microexposure]

	Valid indices (channel numbers) of the returned array are [0..Nchannel-1].
	See noted to sc_getidx().
	
*/
double * sc_getavgflux(signed char isgen) ;

/**
	@brief	Get pointer to the array of second moments of counts in channels in
		[(counts per basetime)^2]
	@param	lag	0: simple dispersion (index<Nchannel) or covariation of counts
	(for returned array index more or equal to Nchannel); 1,2: the same with 
	lag=1,2 counts
	@return	Array of the moment values
	
	Valid indices (channel and combination numbers) are
	[0..NAPCOMB(Nchannel)-1]. See notes to sc_getmean().
*/
double * sc_getsig(int lag) ;

/** Smoothing window width in DESI calculations. Used implicitly in
sc_indices() ) */
#define SC_DESIBIN 3

/** Index writing printf-format in the file with leading "\%s" to insert a
delimiter */
#define SC_IDXFMT "%s%7.4g"
/** Index file delimiter */
#define SC_IDXDLM " "

/** Average flux per microexposure writing printf-format in the file with 
leading "\%s" to insert a delimiter */
#define SC_FLXFMT "%s%7.2f"
/** Average flux file delimiter */
#define SC_FLXDLM " "

/** Moments (means, covariations) writing printf-format in the file with 
leading "\%s" to insert a delimiter */
#define SC_MOMFMT "%s%7.4f"
/** Moment file delimiter */
#define SC_MOMDLM "\t"

/** Not-shifted flag \em isgen in sc_average() */
#define SC_MODENORM 0
/** Shifted flag \em isgen in sc_average() */
#define SC_MODEGEN 1
/** Initial mode: unknown in sc_isgen[] */
#define SC_MODENO -1

/** Maximal lag in calculations of the error of mean for correlated values */
#define SC_MAXLAG 1000

/* Error codes which functions return (used also in WEIF unit): see sc_error()*/
#define SC_ERLO (0)	
/* Fatal errors : programming error */
#define SC_ERNUL (-1)	/**< Error code: NULL pointer supplied */
#define SC_ERMEM (-2)	/**< Error code: MEMORY allocation error */
#define SC_ERNNL (-3)	/**< Error code: Allocation attempt on non-NULL array */
#define SC_ERPAR (-4)	/**< Error code: Non-sense parameter */
#define SC_ERFATAL SC_ERPAR 

/* Non-fatal errors, e.g. by bad input data */
#define SC_ERFIO (-5)	/**< Error code: Bad file info or I/O error */
#define SC_EROFL (-6)	/**< Error code: Not expected count series (index
						storages overfull) */
#define SC_ERNOD (-7)   /**< Error code: no data supplied for calculations */
#define SC_ERZGR (-8)	/**< WEIF: Altitude grids are incompatible */
#define SC_ERSPE (-9)	/**< WEIF: Spectral Energy Distributions are incompatible */
#define SC_ERGEO (-10)	/**< WEIF: Apertures geometry or number are different */
#define SC_ERCRV (-11)  /**< ATMOS: Spectrum and response curve do not match */
#define SC_ERHI (-11)

#endif
